/* Copyright (c) 2015 Siegfried Kettlitz 
 *
 * Implementation of futures.
 * 
 * class future<T> - Virtual base class to wait for and obtain values.
 * class promise<T> - Virtual base class to set values.
 *
 * class future_v<T> - Implementation inheriting the virtual base classes.
 * class future_s<T> - Implementation without inheritance.
 *
 * The value type T stored by the future is templated.
 *
 * Basic usage:
 *
 * // Declare the future with the type e.g. at global scope.
 * future_s<int> fs;
 *
 * // Set the value - e.g. in another thread or in an interrupt.
 * fs.set_value(123);
 *
 * // Wait for the value.
 * fs.wait();
 *
 * // Get the value and reset.
 * fs.get();
 * 
 * // Declare the future with the type e.g. at global scope.
 * future_v<int> fv;
 *
 * // Declare a promise which can be used in a function.
 * promise<int> &pfv = fv;
 *
 * // Set the value.
 * pfv.set_value(123);
 *
 * // Declare a general future which can be used in a function.
 * future<int> &ffv = fv;
 *
 * // Wait for the value.
 * ffv.wait();
 *
 * // Get the value and reset.
 * ffv.get();
 * 
 */
 
// Include guard.
#pragma once


// This function does nothing.
// It is used as a default template parameter.
inline void do_nothing() {}

// Interface of the future.
template< class T >
class future
{
public:
  virtual bool is_ready() = 0;
  virtual T get() = 0;
  virtual void wait() = 0;
};

// Interface of the promise.
template< class T >
class promise
{
public:
  virtual bool accepts_value() = 0;
  virtual void set_value( T ) = 0;
};


/* A simple implementation of a future with virtual inheritance.
 * The template parameter T determines the type of the value.
 * Optionally the functions on_wait and on_error can be specified. 
 */
template< class T, void(on_wait)()=do_nothing, void(on_error)()=do_nothing >
class future_v : public future<T>, public promise<T>
{
  // An enum for the possible states.
  // Empty when created empty or value is read via get().
  // Value_set after the value is set.
  enum class e_status {
    empty, 
    value_set,
  };

  // The status.
  volatile e_status _status;
  
  // The actual value.
  volatile T _value;

public:

  // Start empty.
  future_v() : _status(e_status::value_set) {}

  // Start with value.
  future_v( T value ) : _status(e_status::empty), _value(value) {}

  bool accepts_value()
  {
    // Check the status.
    return _status == e_status::empty;
  }

  bool is_ready() 
  {
    // Check the status.
    return _status == e_status::value_set;
  }

  // Read the value and reset the status back to empty.
  T get() 
  {
    // Sanity check the status.
    if( _status != e_status::value_set ) {
      // Signal an error.
      on_error();
    }

    // Avoid race condition, copy the value.
    T temp = _value;

    // Reset status.
    _status = e_status::empty;
    return temp;
  }

  // Set the value and set the status to value_set.
  void set_value( T value ) 
  {
    if( _status != e_status::empty ) {
      // Signal an error.
      on_error();
    }

    // Set value and change status.
    _value = value;
    _status = e_status::value_set;
  }

  // Loop until the value is set.
  void wait() 
  {
    // Wait until ready.
    while( !is_ready() ) {
      // Idle, optionally do something while waiting e.g. "sleep".
      on_wait();
    }
  }
  
};



/* A simple implementation of a future - no virtual inheritance.
 * The template parameter T determines the type of the value.
 * Optionally the functions on_wait and on_error can be specified. 
 */
template< class T, void(on_wait)()=do_nothing, void(on_error)()=do_nothing >
class future_s
{
  // An enum for the possible states.
  // Empty when created empty or value is read via get().
  // Value_set after the value is set.
  enum class e_status {
    empty, 
    value_set,
  };

  // The status.
  volatile e_status _status;
  
  // The actual value.
  volatile T _value;

public:

  // Start empty.
  future_s() : _status(e_status::value_set) {}

  // Start with value.
  future_s( T value ) : _status(e_status::empty), _value(value) {}

  bool accepts_value()
  {
    // Check the status.
    return _status == e_status::empty;
  }

  bool is_ready() 
  {
    // Check the status.
    return _status == e_status::value_set;
  }

  // Read the value and reset the status back to empty.
  T get() 
  {
    // Sanity check the status.
    if( _status != e_status::value_set ) {
      // Signal an error.
      on_error();
    }

    // Avoid race condition, copy the value.
    T temp = _value;

    // Reset status.
    _status = e_status::empty;
    return temp;
  }

  // Set the value and set the status to value_set.
  void set_value( T value ) 
  {
    if( _status != e_status::empty ) {
      // Signal an error.
      on_error();
    }

    // Set value and change status.
    _value = value;
    _status = e_status::value_set;
  }

  // Loop until the value is set.
  void wait() 
  {
    // Wait until ready.
    while( !is_ready() ) {
      // Idle, optionally do something while waiting e.g. "sleep".
      on_wait();
    }
  }
  
};